use io;
use random;

/* Define input data */
function input() {
    nbPeriods = 3;
    prices = { 100, 300, 400 };
    meanDemands = { 50, 20, 30 };
    purchasePrice = 80;
    evaluatedPoints = {{
        "point": { 100, 50, 30 },
        "value": 4740.99
    }};
    nbSimulations = round(1e6);
    seed = 1;

    // Create random module
    rdm = random.create();
}

// External function
function revenueManagement(v1, v2, v3) {
    // Initial quantity purchased
    nbUnitsPurchased = v1;
    // Number of units that should be left for future periods
    nbUnitsReserved = { v2, v3, 0 };

    // Set seed for reproducibility
    rdm.init(seed);
    // Create distribution
    rateParam = 1.0;
    scaleParam = 1.0;
    X[i in 0...nbSimulations] = gammaSample(scaleParam);
    Y[i in 0...nbSimulations][j in 0...nbPeriods] = exponentialSample(rateParam);

    // Run simulations
    sumProfit = 0.0;
    for [i in 0...nbSimulations] {
        remainingCapacity = nbUnitsPurchased;
        for [j in 0...nbPeriods] {
            // Generate demand for period j
            demandJ = round(meanDemands[j] * X[i] * Y[i][j]);
            nbUnitsSold = min(max(remainingCapacity - nbUnitsReserved[j], 0),
                    demandJ);
            remainingCapacity -= nbUnitsSold;
            sumProfit += prices[j] * nbUnitsSold;
        }
    }

    // Calculate mean revenue
    meanProfit = sumProfit / nbSimulations;
    meanRevenue = meanProfit - purchasePrice * nbUnitsPurchased;

    return meanRevenue;
}

function exponentialSample(rateParam) {
    u = rdm.nextUniform(0, 1);
    return log(1 - u) / (-rateParam);
}

function gammaSample(scaleParam) {
    return exponentialSample(scaleParam);
}

/* Declare the optimization model */
function model() {
    // Declare decision variables
    variables[i in 0...nbPeriods] <- int(0, 100);

    // Create the function
    funcExpr <- doubleExternalFunction(revenueManagement);
    // Call function
    funcArgs[0] <- funcExpr;
    funcArgs[i in 1...nbPeriods+1] <- variables[i-1];
    funcCall <- call[arg in funcArgs](arg);

    // Declare constraints
    for [i in 1...nbPeriods]
        constraint variables[i] <= variables[i - 1];

    // Maximize function call
    maximize funcCall;

    // Enable surrogate modeling
    context = funcExpr.context;
    surrogateParams = context.enableSurrogateModeling();

    // Set lower bound
    context.lowerBound = 0.0;
}

function param() {
    // Set the maximum number of evaluations
    if (evaluationLimit == nil) surrogateParams.evaluationLimit = 30;
    else surrogateParams.evaluationLimit = evaluationLimit;

    // Add evaluation points
    for [k in 0...count(evaluatedPoints)] {
        evaluationPoint = surrogateParams.createEvaluationPoint();
        for [i in 0...nbPeriods]
            evaluationPoint.addArgument(evaluatedPoints[k]["point"][i]);
        evaluationPoint.returnValue = evaluatedPoints[k]["value"];
    }
}

/* Write the solution in a file */
function output() {
    if (solFileName != nil) {
        local solFile = io.openWrite(solFileName);
        solFile.println("obj=", funcCall.value);
        solFile.println("b=", variables[0].value);
        for [i in 1...nbPeriods]
            solFile.println("r", i + 1, "=", variables[i].value);
    }
}
